#include <afxwin.h>

#include "CEngineDef.h"
#include "CEngineDefinitions.h"
#include "pdn.h"


char* CPDN::pdn_tag[9] = 
	{ "Event", "Site", "Date", "Round", "White", "Black", "Result", "SetUp", "FEN" } ;

const char* CPDN::str_result[GAMEHEADER_RESULT_NUMBER] =	
	{ "1-0", "0-1", "1/2-1/2", "*", "?", "2-0", "0-2", "1-1", "3-0", "0-3", "" } ;

const int CPDN::num_result[GAMEHEADER_RESULT_NUMBER] =
	{ 1, 2, 3, 4, 4, 1, 2, 3, 1, 2, 4 } ;

CPDN::CPDN(void)
{

	pdn_val[0] = pdn_event ;
	pdn_val[1] = pdn_site ;
	pdn_val[2] = pdn_date ;
	pdn_val[3] = pdn_round ;
	pdn_val[4] = pdn_white ;
	pdn_val[5] = pdn_black ;
	pdn_val[6] = pdn_result ;
	pdn_val[7] = pdn_setup ;	
	pdn_val[8] = pdn_fen ;

}


CPDN::~CPDN(void)
{
}


void CPDN::Write_PDNGame(PENGINEDATA pEngineData)
{
	int i, iMoveIndex, iNGames, iNMoves, iResult ;
	PDXPMATCH pDXPMatch ;
	CString csTAGInfo, csWhite, csBlack, csMove, csMoveList, csMoveIndex, csResult ;
	CFile* pDXPFile ;

	pDXPFile	= &(pEngineData->DXPMatch.MatchFile) ;
	pDXPMatch	= &(pEngineData->DXPMatch) ;

	if ( pDXPMatch->bDXPClientTurn ) {
		csWhite = pDXPMatch->csClientName ;
		csBlack = pDXPMatch->csServerName ;
	} else {
		csWhite = pDXPMatch->csServerName ;
		csBlack = pDXPMatch->csClientName ;
	}

	iNGames = pDXPMatch->iNGames ;
	iResult = pDXPMatch->iResult[iNGames-1] ;

	if ( iResult == DXP_WIN )
		csResult = "2-0" ;
	else if ( iResult == DXP_DRAW )
		csResult = "1-1" ;
	else if ( iResult == DXP_LOSE )
		csResult = "0-2" ;
	else
		csResult = "? " ;

	CArchive CArchive_DXPFile(pDXPFile, CArchive::store) ;
	
	// Write Event
	csTAGInfo.Format("[%s %cGame %d, opening %s %s%c]\n", PDN_EVENT, 34, pDXPMatch->iNGames, pDXPMatch->csBallotWhiteMove, pDXPMatch->csBallotBlackMove, 34) ;
	CArchive_DXPFile.WriteString(csTAGInfo) ;

	// Write White Player
	csTAGInfo.Format("[%s %c%s%c]\n", PDN_WHITE, 34, csWhite, 34) ;
	CArchive_DXPFile.WriteString(csTAGInfo) ;

	// Write Black Player
	csTAGInfo.Format("[%s %c%s%c]\n", PDN_BLACK, 34, csBlack, 34) ;
	CArchive_DXPFile.WriteString(csTAGInfo) ;

	// Write Result
	csTAGInfo.Format("[%s %c%s%c]\n\n", PDN_RESULT, 34, csResult, 34) ;
	CArchive_DXPFile.WriteString(csTAGInfo) ;

	iNMoves = pEngineData->iGameHistoryActualMove ;

	csMoveList.Empty() ;

	for ( i=0; i<iNMoves; i++ ) {

		csMove = pEngineData->GameHistory[i].csMove ;

		iMoveIndex = (i+2)/2 ;

		if ( (i+1)%2 ) {
			csMoveIndex.Format("%d. ", iMoveIndex ) ;
			csMoveList += csMoveIndex ;
		}

		csMoveList += csMove ;
		csMoveList += " " ;

//		if ( !( (i+1)%10 ) )
//			csMoveList += "\n" ;
//		else
//			csMoveList += " " ; 
	}

	csResult += "\n\n" ;

	// Write Moves 
	CArchive_DXPFile.WriteString(csMoveList) ;

	// Write (again) Result
	CArchive_DXPFile.WriteString(csResult) ;

	CArchive_DXPFile.Close() ;
}


void CPDN::strip_trailing(char *ptr)
{
	int x ;

	x = (int)strlen(ptr) ;
	do {
		ptr[x--] = '\0' ;
	} while ( x >= 0 && (ptr[x] == ' ' || ptr[x] == '\r' || ptr[x] == '\n' ) ) ;
}


char CPDN::read_char(void)
{
	char ch ;

	if ( pdn_ptr >= pdn_text + pdn_size )
		return '\0' ;
    
	ch = *pdn_ptr++ ;
    
	if ( ch == '\0' )
		ch = ' ' ;
    
	if ( ch == '\r' || ch == '\n' )
		pdn_pos = 0 ;
	else
		pdn_pos++ ;

    return ch ;
}


void CPDN::put_char(void)
{
    
	pdn_ptr-- ;

	if ( pdn_pos > 0 )
		pdn_pos-- ;
}


bool CPDN::is_white(char ch)
{

    switch ( ch ) {
    
		case ' ' :
		case ',' :
		case '\t' :
		case '\n' :
		case '\f' :
		case '\r' :
			return true ;
    
		default:
			break ;
    }

    return false ;
}


bool CPDN::is_symbol(char ch)
{
	int file_type = 2 ;		// PDN File !

	if (ch >= 'A' && ch <= 'Z')
		return true ;

	if (ch >= 'a' && ch <= 'z')
		return true ;

	if (ch >= '0' && ch <= '9')
		return true ;

    switch (ch) {

		case '_' :
		case '+' :
		case '#' :
		case '=' :
		case ':' :
		case '-' :
		case '/' :
			return true ;

		case '*':
			return ( file_type == 3 ) ;
    
		default:
			break ;
	}

	return false ;
}


bool CPDN::Read_Token(void)
{
    
	char ch ;
	int  i ;
	
	i = 0 ;
	pdn_token[0] = '\0' ;
	
	do {							// skip white space
		ch = read_char() ;
	} while ( is_white(ch) ) ;

	if ( ch == '\0' )
		return false ;

	if ( ch == '"' ) {				// string token

		do {
            
			ch = read_char() ;
            
			switch ( ch ) {

				case '\0':				// end of string
				case '"':
				case '\r':
				case '\n':
					break ;

				case '\\':				// quote character
					ch = read_char() ;
					if ( ch == '\0' || ch == '\r' || ch == '\n' )
						break ;
                
					if ( i < PDNLEN - 1 )
						pdn_token[i++] = ch ;
					continue ;

				default:				// regular character
					if (i < PDNLEN - 1)
						pdn_token[i++] = ch ;
					continue ;
            }

            break;
 
		} while(true);

	} else if ( ch == ';' || (pdn_pos == 1 && ch == '%') ) {	// comment to end of line

		do {
			ch = read_char() ;
		} while ( ch != '\0' && ch != '\r' && ch != '\n' ) ;
        
		return Read_Token() ;

	} else if ( ch == '{' )  {				// comment in braces, not recursive

		do {
			ch = read_char() ;
		} while ( ch != '\0' && ch != '}' ) ;
        
		return Read_Token() ;

	} else if ( ch == '(' )	{				// recursive annotation variations

		do {
            if ( !Read_Token() )
                return false ;
        } while ( strcmp(pdn_token, ")") != 0 ) ;
        return Read_Token();

	} else if (ch == '<') {					// reserved <>, assume recursive

		do {
			if ( !Read_Token() )
				return false ;
		} while ( strcmp(pdn_token, ">") != 0 ) ;
        
		return Read_Token() ;

	} else if (ch == '$') {					// NAG token

        do {
			ch = read_char() ;
		} while (ch >= '0' && ch <= '9') ;

        put_char() ;
        return Read_Token() ;

	} else {                                // regular token

		pdn_token[i++] = ch ;
		while ( is_symbol(ch) ) {

            ch = read_char() ;

			if ( ch == ' ' && ( pdn_token[i-1] == '-' || pdn_token[i-1] == 'x' ) )	// 2nd condition added to detect move aa- b aax b !!
				ch = '0' ;			

			if ( is_symbol(ch) ) {

				if ( i < PDNLEN - 1 )
					pdn_token[i++] = ch ;

			} else
				put_char() ;
		}
	}
    
	pdn_token[i] = '\0' ;
    
	return true ;
}


bool CPDN::Read_Header(void)
{
	return ( Read_PDNHeader() ) ;
}


bool CPDN::Read_FEN(void)
{
	char *fen ;
	int  i, iColor, iPiece, iPos ;
    
	if ( pdn_fen[0] != '\0' ) {					// FEN available 
        
		for ( i=1 ; i<50 ; i++ )
			m_iBoard[i] = EMPTY ;

		CharLower(pdn_fen) ;
		fen = pdn_fen ;
        
		while( is_white(*fen) )
			fen++;
        
		if ( *fen == '\0' )
            return false ;

		if ( *fen == 'w' )						// Side to move
			m_bTurn = true ;					// White to move
		else if ( *fen == 'b' )
			m_bTurn = false ;					// Black to move
		else
			return false ;						// Illegal value !

        do {									// Colon for piece list
            
			while( *fen != ':' && *fen != '\0' )
				fen++ ;
 
			if ( *fen == '\0' )
				return false ;

			fen++ ;
            
			while( is_white(*fen) )
				fen++ ;
            
			if ( *fen == '\0' )
				return false ;
            
			if ( *fen == 'w' )					// Color of piece list
				iColor = 1 ;
			else if ( *fen == 'b' )
				iColor = 0 ;
			else
				return false ;
            
			fen++ ;
            
			do {
                
				while( is_white(*fen) )
					fen++ ;

				if ( *fen == '\0' )
					return false ;
                
				if ( *fen == 'k' ) {			// Indication of king

					if ( iColor == 1 )
						iPiece = WHITEKING ;
					else
						iPiece = BLACKKING ;

					fen++ ;
                    
					while( is_white(*fen) )
						fen++ ;
                    
					if ( *fen == '\0' )
						return false ;
                
				} else {

					if ( iColor == 1 )
						iPiece = WHITEMAN ;
					else
						iPiece = BLACKMAN ;
				}

				iPos = 0 ;						// Get square number
                
				while ( *fen >= '0' && *fen <= '9' )
					iPos = 10 * iPos + *fen++ - '0' ;

                if ( iPos >= 1 && iPos <= 50 ) {

					if ( iPos <= 5 && iPiece == WHITEMAN ) 
						iPiece = WHITEKING ;	// Promote white man
						
					if 	( iPos >= 46 && iPiece == BLACKMAN )
						iPiece = BLACKKING ;	// Promote man on last row
                    
					m_iBoard[iPos] = iPiece ;	// Place piece on the board
				}
            
			} while ( is_white(*fen) ) ;		// End of a square number
        
		} while ( *fen == ':' ) ;				// End of a color's piece list
    
	} else {

		m_bTurn = true ;						// No FEN, set beginning position */
    
		for ( i=1 ; i<=20 ; i++ )
			m_iBoard[i] = BLACKMAN ;

		for ( i=21; i<=30 ; i++ )
			m_iBoard[i] = EMPTY ;

		for ( i=31 ; i<=50 ; i++ )
			m_iBoard[i] = WHITEMAN ;
	}

	return ( true ) ;
}


bool CPDN::Read_Moves(void)
{
	char *startptr, *ptr ;
	int  iIndex, pos, m ;

	m_iMove = 0 ;

	for ( m = 0; m < elements(str_result) - 1; m++ )
        if ( strcmp(pdn_result, str_result[m]) == 0 )
            break ;
    
	pdn_numresult = num_result[m] ;		// convert result string to numeric
	startptr = pdn_ptr ;
	pdn_ptr = startptr ;
		
	while (pdn_token[0] != '\0' && strcmp(pdn_token, pdn_result) != 0 && strcmp(pdn_token, "[") != 0 ) {

		// stop at any of the standard or not-so-standard result codes
            
		for (m = 0; m < elements(str_result) - 1; m++)
			if ( strcmp(pdn_token, str_result[m]) == 0 )
				break ;

		if ( m < elements(str_result) - 1 )
		break ;

		ptr = pdn_token ;			// add numbers in token to select list
		iIndex = 0 ;

		do {
			while ( *ptr != '\0' && (*ptr < '0' || *ptr > '9') )
				ptr++ ;

			
			if (*ptr != '\0') {
				pos = 0 ;
				
				while ( *ptr >= '0' && *ptr <= '9' )
					pos = 10*pos + *ptr++ - '0' ;

				m_iMovePDN[m_iMove][++iIndex] = pos ;
			}

		} while ( *ptr != '\0' ) ;

		m_iMovePDN[m_iMove][0] = iIndex;

		if ( Read_Token() && strcmp(pdn_token, ".") == 0 ) {

			if ( !Read_Token() )	// if followed by period, it's a move number
				break ;

			continue ;			// go on with token following period 
		}

		++m_iMove ;
	}

	return ( true ) ;
}


int CPDN::Read_Game(int iPDNGame, PGAMEHEADER pPDNGameHeader)
{
	bool b ;
	int i, iResult, iYear, iMonth, iDay ;

	pdn_last		= pdn_ptr = pdn_top = pdn_text ;	
	pdn_token[0]	= '\0' ;	
	pdn_pos			= 0 ;
   
	i = 0 ;   
	do {
		pdn_last = pdn_top ;
		b = Read_Header() ;
		i++ ;
	} while ( b && i < iPDNGame ) ;

	if ( !b || i < iPDNGame ) {
		iPDNGame = i - 1;
		pdn_ptr = pdn_top = pdn_last ;
		pdn_token[0] = '\0';
		pdn_pos = 0 ;
		Read_Header() ;
	}

	pPDNGameHeader->csWhite      =  pdn_val[4] ;
	pPDNGameHeader->csBlack      =  pdn_val[5] ;
	pPDNGameHeader->csTournament =  pdn_val[0] ;
	pPDNGameHeader->csSite       =  pdn_val[1] ;

	iResult = 3 ;
	for ( i=0; i<GAMEHEADER_RESULT_NUMBER; i++ )
		if ( strcmp(pdn_val[6], str_result[i] ) == 0 )
			iResult = i ;

	pPDNGameHeader->iResult = num_result[iResult] - 1 ;

	if ( strlen(pdn_val[2]) ) {
		pPDNGameHeader->iCDate = 1 ;
		sscanf(pdn_val[2],"%d.%d.%d",&iYear, &iMonth, &iDay) ;
		pPDNGameHeader->iYear  = iYear ;
		pPDNGameHeader->iMonth = iMonth ; 
		pPDNGameHeader->iDay   = iDay ;
	} else
		pPDNGameHeader->iCDate = 0 ;

	if ( strlen(pdn_val[3]) ) {
		pPDNGameHeader->iCRound = 1 ;
		sscanf(pdn_val[3],"%d",&pPDNGameHeader->iRound) ;
	} else {
		pPDNGameHeader->iCRound = 0 ;
		pPDNGameHeader->iRound  = 0 ;
	}

	Read_FEN() ;
	Read_Moves() ;
        
	if ( pdn_ptr != pdn_top && pdn_ptr[-1] == '[' )
		pdn_ptr-- ;            // step back from start of next game

	*pdn_ptr = '\0';

	return ( m_iMove ) ;
}


bool CPDN::Read_PDNHeader(void)
{
	int i ;

	while ( pdn_token[0] != '[' ) {						// find first tag section
		if ( !Read_Token() )
			return false ;
	}

    for ( i = 0; i < elements(pdn_val); i++)
		memset(pdn_val[i], 0, PDNLEN) ;					// clear all tag values
	    
	pdn_top = pdn_ptr - 1 ;								// start of this game's PDN text

    do {
		if ( !Read_Token() )							// read tag
            return false ;

        for (i = 0; i < elements(pdn_val); i++) {		// case insensitive compare
 
			if ( _stricmp(pdn_token, pdn_tag[i]) == 0 ) {
                if ( !Read_Token() )					// next token is tag value
					return false;
                
				strcpy(pdn_val[i], pdn_token) ;
				break ;
			}
		}
        
		do {
			if ( !Read_Token() )							// find end bracket
				return false ;
        } while ( strcmp(pdn_token, "]") != 0 ) ;

		if ( !Read_Token() )								// possible subsequent start bracket */
			return false ;

	} while ( strcmp(pdn_token, "[") == 0 ) ;
    
	return ( true ) ;
}


int CPDN::Read_PDN(char* pcPDN, int iFileSize)
{
	int iNPDNGame ;

	iNPDNGame = 0 ;

	pdn_text = pcPDN ;
	pdn_size = iFileSize ;

	pdn_last = pdn_ptr = pdn_top = pdn_text ;

	pdn_token[0]	= '\0' ;
	pdn_pos			= 0 ;

	while ( Read_PDNHeader() )	// Determine Nr. Games in PDN file 
		++iNPDNGame ;

	return ( iNPDNGame ) ;
}